#include "encoding.hpp"
#include <stdint.h>
#include <string.h>

#include <QRgb>
#include <QColor>

/***
	Standard tools.
***/

static int color(QRgb rgb)
{
	QColor c(rgb);
	return c.red() + c.green() + c.blue() < 384;
}

static const uint8_t *encode_ml(QImage image)
{
	int real_width = ((image.width() + 7) >> 3) << 3;
	int bytes = (real_width >> 3) * image.height();
	uint8_t cell = 0;
	uint8_t *data;
	int x, y;
	int n = 0;

	data = (uint8_t *)calloc(bytes, 1);
	if(!data) return NULL;

	for(y = 0; y < image.height(); y++)
	{
		for(x = 0; x < real_width; x++)
		{
			cell <<= 1;
			if(x < image.width())
				cell |= color(image.pixel(x, y));

			if((x & 7) == 7)
			{
				data[n++] = cell;
				cell = 0;
			}
		}
	}

	return data;
}

static const uint8_t *encode_variant(QImage image)
{
	return NULL;
}

/***
	Image encoding.
***/

int encode_image(QImage image, struct Image *ml_code, struct Image *variant)
{
	int width = image.width(), height = image.height();
	if(!width || !height) return 1;

	ml_code->size = ((width + 7) >> 3) * height;
	ml_code->data = encode_ml(image);

	variant->size = (width * height + 7) >> 3;
	variant->data = encode_variant(image);

	if(!ml_code->data) return -1;
	return 0;
}

/***
	Set encoding.
***/

int encode_set(QImage image, int width, int height, enum BorderKind border,
	struct Set *ml_code, struct Set *variant)
{
	int el_hor, el_ver;
	QImage element;
	int i, j;
	int x, y;

	// Determinin the number of elements in the set.
	switch(border)
	{
	case Border_None:
		el_hor = image.width() / width;
		el_ver = image.height() / height;
		break;
	case Border_Separate:
		el_hor = (image.width() + 1) / (width + 1);
		el_ver = (image.height() + 1) / (height + 1);
		break;
	case Border_Full:
	default:
		el_hor = (image.width() - 1) / (width + 1);
		el_ver = (image.height() - 1) / (height + 1);
		break;
	}

	if(!el_hor || !el_ver) return 1;

	ml_code->elements = el_hor * el_ver;
	ml_code->element_size = ((width + 7) >> 3) * height;
	ml_code->element_width = width;
	ml_code->element_height = height;
	ml_code->data = (const uint8_t **)malloc(ml_code->elements *
		sizeof(const uint8_t *));

	variant->elements = el_hor * el_ver;
	variant->element_size = (width * height + 7) >> 3;
	variant->element_width = width;
	variant->element_height = height;
	variant->data = NULL;

	if(!ml_code->data) return -1;

	for(i = 0; i < el_ver; i++) for(j = 0; j < el_hor; j++)
	{
		x = (width + (border == Border_Separate)) * j
			+ (border == Border_Full);
		y = (height + (border == Border_Separate)) * i
			+ (border == Border_Full);

		element = image.copy(x, y, width, height);
		ml_code->data[i * el_hor + j] = encode_ml(element);
	}

	return 0;
}

/***
	Map encoding using ML tileset as reference.
***/

int encode_map(QImage image, struct Set *ml_set, struct Map *map)
{
	int width = ml_set->element_width, height = ml_set->element_height;
	const uint8_t *ml_tile;
	int el_hor, el_ver;
	uint8_t *map_data;
	int warning = 0;
	QImage tile;
	int i, j, n;

	el_hor = image.width() / width;
	el_ver = image.height() / height;
	if(!el_hor || !el_ver) return 1;

	map->elements = el_hor * el_ver;
	map->width = el_hor;
	map->height = el_ver;
	map_data = (uint8_t *)malloc(width * height);
	map->data = (const uint8_t *)map_data;
	if(!map->data) return 0;

	for(i = 0; i < el_ver; i++) for(j = 0; j < el_hor; j++)
	{
		tile = image.copy(j * width, i * height, width, height);
		ml_tile = encode_ml(tile);

		for(n = 0; n < ml_set->elements; n++)
		{
			if(!memcmp(ml_set->data[n], ml_tile,
				ml_set->element_size)) break;
		}

		if(n < ml_set->elements) map_data[i * el_hor + j] = n;
		else map_data[i * el_hor + j] = 0xff, warning = 1;

		free((void *)ml_tile);
	}

	return warning ? 2 : 0;
}
